--
--	cache.vhd
--
--	Bytecode caching
--
--
--	2005-01-11	first version
--

Library IEEE;
use IEEE.std_logic_1164.all;
use ieee.numeric_std.all;

entity cache is
generic (jpc_width : integer);

port (

	clk, reset	: in std_logic;

	bc_len		: in std_logic_vector(9 downto 0);		-- length of method in words
	bc_addr		: in std_logic_vector(17 downto 0);		-- memory address of bytecode

	find		: in std_logic;							-- start lookup

	-- start of method in bc cache
	-- in 32 bit words - we load only at word boundries
	bcstart		: out std_logic_vector(jpc_width-3 downto 0);

	rdy			: out std_logic;						-- lookup finished
	in_cache	: out std_logic							-- method is in cache

);
end cache;

architecture rtl of cache is

--
--	signals for mem interface
--
	type state_type		is (
							idle, s1, s2
						);
	signal state 		: state_type;

	signal block_addr	: std_logic_vector(jpc_width-3 downto 0);
	signal bla, blb		: std_logic_vector(17 downto 0);
	signal blc, bld		: std_logic_vector(17 downto 0);
	signal nxt			: unsigned(1 downto 0);

begin

	bcstart <= block_addr;

process(clk, reset, find)

begin
	if (reset='1') then
		state <= idle;
		rdy <= '1';
		in_cache <= '0';
		block_addr <= (others => '0');
		nxt <= "00";
		bla <= (others => '0');
		blb <= (others => '0');
		blc <= (others => '0');
		bld <= (others => '0');

	elsif rising_edge(clk) then

		case state is

			when idle =>
				state <= idle;
				rdy <= '1';
				if find = '1' then
					rdy <= '0';
					state <= s1;
				end if;

			-- check for a hit
			when s1 =>
				in_cache <= '0';
				block_addr <= std_logic_vector(nxt) & "000000";
				state <= s2;

				if bla = bc_addr then
					block_addr <= "00000000";
					in_cache <= '1';
					state <= idle;
				end if;
				if blb = bc_addr then
					block_addr <= "01000000";
					in_cache <= '1';
					state <= idle;
				end if;
				if blc = bc_addr then
					block_addr <= "10000000";
					in_cache <= '1';
					state <= idle;
				end if;
				if bld = bc_addr then
					block_addr <= "11000000";
					in_cache <= '1';
					state <= idle;
				end if;

			-- correct tag memory on a miss
			when s2 =>
				if nxt = "00" then

					bla <= bc_addr;
					if bc_len(7 downto 6) = "01" then
						blb <= (others => '0');
					elsif bc_len(7 downto 6) = "10" then
						blb <= (others => '0');
						blc <= (others => '0');
					elsif bc_len(7 downto 6) = "11" then
						blb <= (others => '0');
						blc <= (others => '0');
						bld <= (others => '0');
					end if;

				elsif nxt = "01" then

					blb <= bc_addr;
					if bc_len(7 downto 6) = "01" then
						blc <= (others => '0');
					elsif bc_len(7 downto 6) = "10" then
						blc <= (others => '0');
						bld <= (others => '0');
					elsif bc_len(7 downto 6) = "11" then
						blc <= (others => '0');
						bld <= (others => '0');
						bla <= (others => '0');
					end if;

				elsif nxt = "10" then

					blc <= bc_addr;
					if bc_len(7 downto 6) = "01" then
						bld <= (others => '0');
					elsif bc_len(7 downto 6) = "10" then
						bld <= (others => '0');
						bla <= (others => '0');
					elsif bc_len(7 downto 6) = "11" then
						bld <= (others => '0');
						bla <= (others => '0');
						blb <= (others => '0');
					end if;

				elsif nxt = "11" then

					bld <= bc_addr;
					if bc_len(7 downto 6) = "01" then
						bla <= (others => '0');
					elsif bc_len(7 downto 6) = "10" then
						bla <= (others => '0');
						blb <= (others => '0');
					elsif bc_len(7 downto 6) = "11" then
						bla <= (others => '0');
						blb <= (others => '0');
						blc <= (others => '0');
					end if;

				end if;

				state <= idle;
				-- don't increment nxt for too short methods
				if to_integer(signed(bc_len)) > 2 then
					nxt <= nxt + unsigned(bc_len(7 downto 6)) + 1;
				end if;


		end case;
					
	end if;
end process;

end rtl;
